import math
from copy import copy, deepcopy
from polysurface import spline_double_divide, Offset, SurfaceFromCurves
from car_util import * # yeahhh don't do this
from pyopenscad import * # or this.  *sigh*.
from roof import roof_rear_pts, Roof
from bonnet import Bonnet
from front_section import Front
from floor import Floor
from airbox import Airbox
from backbox import BackBox
from rearsection import RearSection
from door_util import (calc_door_offsets, pmid, doorlinepts, doorsillpts,
                       make_points)

###############
# dashboard and doorholder
###############

def dashboard_door(airbox_pts, bpoints, roofpts, cfloor_pts, fpoints,
                   floor_offs):

    t = Group(name="dashboard_door")
    aback, atop = airbox_pts
    bbase, bscoop = bpoints
    rbase, rmid, rscreen, doorpts = roofpts

    # dashboard
    t.add_pipe(aback[1][-1], doorpts[2], dthin, AUTO, jlbrot2)
    t.add_pipe(aback[1][-1], doorpts[3], dthin, jrb)
    # not now - no joints
    t.add_doorline_pipe(7, 6, doorpts[2], doorpts[3])
    # XXX if faulty above t.add_pipe(doorpts[2], doorpts[3])

    t.add_pipe(aback[0][-1], doorpts[1], dthin, AUTO, jrbrot2)
    t.add_pipe(aback[0][-1], doorpts[0], dthin, jlb)
    t.add_doorline_pipe(7, 6, doorpts[1], doorpts[0])
    # XXX if faulty above t.add_pipe(doorpts[1], doorpts[0], dthin, jdown)

    # bracing to bonnet
    t.add_pipe(doorpts[2], bscoop[1][0], dthin, jlfd3, jright)
    t.add_pipe(doorpts[3], bscoop[1][0])
    t.add_pipe(aback[1][-1], bbase[3][0])
    t.add_pipe(bbase[3][0], doorpts[2], dthin, AUTO, jfd)
    t.add_pipe(doorpts[3], bbase[3][0])

    t.add_pipe(doorpts[1], bscoop[0][0], dthin, jrfd3, jleft)
    t.add_pipe(doorpts[0], bscoop[0][0])
    t.add_pipe(aback[0][-1], bbase[0][0])
    t.add_pipe(bbase[0][0], doorpts[1], dthin, AUTO, jfd)
    t.add_pipe(doorpts[0], bbase[0][0])

    # door line left
    t.add_doorline_pipe(0, 1, doorpts[3], cfloor_pts[-1][-1])
    # XXX if faulty above t.add_pipe(doorpts[3], cfloor_pts[-1][-1])
    t.add_pipe(bbase[3][0], cfloor_pts[-1][-1], dthin, jbd, jfu)

    # door line right
    t.add_pipe(doorpts[0], cfloor_pts[0][-1])
    t.add_pipe(bbase[0][0], cfloor_pts[0][-1], dthin, jbd, jfu)

    return t

###########
# door sill
###########

def door_sill_pts(t, airbox_pts, bpoints, roofpts, cfloor_pts, fpoints,
                   rearbox_pts, floor_offs):

    aback, atop = airbox_pts
    bbase, bscoop = bpoints
    rbase, rmid, rscreen, doorpts = roofpts
    roof, boxbase, midbox, midfront = rearbox_pts

    offs = 0.019
    rad = 0.023
    # XXX nope, too close pdb = pmid(rscreen[-1], rbase[0], doorpts[2], rad)
    pdb = rotate_x(rotate_z([0,rad,0], -25), 50)
    pdf = pmid(rbase[0], midfront[3], rscreen[-1], rad, True)
    pfwd = pmid(midfront[3], cfloor_pts[3][0], rbase[0], rad, True)
    puf = pmid(cfloor_pts[3][0], cfloor_pts[3][-1], midfront[3], rad, True)
    pub = pmid(cfloor_pts[3][-1], cfloor_pts[3][0], doorpts[3], rad)
    pback = pmid(doorpts[3], doorpts[2], cfloor_pts[3][-1], -rad)
    pbd2 = pmid(doorpts[2], rscreen[-1], doorpts[3], rad)

    doorlinepts.append(doorpts[3])
    doorlinepts.append(cfloor_pts[3][-1])
    doorlinepts.append(cfloor_pts[3][0])
    doorlinepts.append(midfront[3])
    doorlinepts.append(rbase[0])
    doorlinepts.append(rscreen[-1])
    doorlinepts.append(doorpts[2])

    # ok so we go round the frame, with offsets pulling in, one at a time
    for i in range(len(doorlinepts)):
        dist = rad
        inv = False
        if i in [2,3,4]:
            inv = True
        if i == 0:
            dist = -dist
        if i == 5:
            poff = rotate_x(rotate_z([0,rad,0], -25), 50)
        elif i == 3:
            poff = rotate_x(rotate_z([0,-rad,0], 25), -30)
        else:
            p1 = doorlinepts[i]
            p2 = doorlinepts[(i+1) % 7]
            p3 = doorlinepts[(i+6) % 7]
            poff = pmid(p1, p2, p3, dist, inv)
        doorsillpts.append(padd(doorlinepts[i], poff))
    for pos in doorsillpts:
        print "pos", pos
        t.add_joint(pos)

    return [doorsillpts]

def door_sill(airbox_pts, bpoints, roofpts, cfloor_pts, fpoints,
                   rearbox_pts, floor_offs):

    t = Group(name="door_sill")
    door_pts = door_sill_pts(t, airbox_pts, bpoints, roofpts, cfloor_pts,
                              fpoints, rearbox_pts, floor_offs)
    aback, atop = airbox_pts
    bbase, bscoop = bpoints
    rbase, rmid, rscreen, doorpts = roofpts
    roof, boxbase, midbox, midfront = rearbox_pts

    for row in door_pts:
        for i in range(-1, len(row)-2):
            i1 = (i + 7) % 7
            i2 = (i + 1) % 7
            offs1, offs2 = calc_door_offsets(i)
            t.add_pipe(row[i1], row[i2], dthin, offs1, offs2, rot=False)

    # XXX hmmm.... this is in the way (reduces driver visibility)
    #rbase, rmid, rscreen, doorpts = roofpts
    #t.add_pipe(doorpts[3], row[-1], diameter=0.018) # see doorframe

    return door_pts, t

################
# door
################

def door_pts(t, roofpts, doorsill_pts, floor_offs):

    rbase, rmid, rscreen, doorpts = roofpts
    doorsill = doorsill_pts[0]
    rp = doorpts[2]

    door_pts = []

    offs = 0.02
    pdown, pup, pfwd, pback, pdb, pdf, puf, pub = make_points(offs, 0.00)

    # inner part
    p = []
    p.append(padd(doorsill[0], pback))
    p.append(padd(doorsill[1], pub))
    p.append(padd(doorsill[2], puf))
    p.append(padd(doorsill[3], pfwd))
    p.append(padd(doorsill[4], pback))#[pdf[0],0.011, -0.001]))
    wininset = 0.026
    #pdb[1] -= wininset # top window at roofline, move inwards
    pdscrn = [0.0235, -0.0055, 0.0105]
    p.append(padd(doorsill[5], pdb))#pdscrn))
    for i, pos in enumerate(p):
        xangle = 0.0
        zangle = 0.0
        yangle = 0.0
        #if i == 5:
            #xangle = 15.0
        if i in [4, 5]:
            yangle = -20.0
            zangle = 3.5
        t.add_joint(pos, xangle=xangle, yangle=yangle, zangle=zangle)
    door_pts.append(p)
    inner = p
    pdb2 = pdb

    # outer part
    offs = -0.025
    pdown, pup, pfwd, pback, pdb, pdf, puf, pub = make_points(offs, -1.0)
    p = []
    p.append(padd(inner[0], pback))
    p.append(padd(inner[1], pub))
    p.append(padd(inner[2], puf))
    p.append(padd(inner[3], pfwd))
    p.append(padd(inner[3], pdf)) # skipped
    p.append(padd(inner[5], pdb)) # skipped
    #[-0.00, 0.045-wininset, -0.01])) # window, at screen-door
    p.append(padd(doorsill[6], pdb2))
    # front corner of door
    p.append(padd(rp, puf))
    # add 3 sets of bars, bit of fudging here
    p.append(padd(p[0], [0.01, -0.01, -0.15]))
    p.append(padd(p[0], [0.03, -0.03, 0.02]))
    p.append(padd(p[0], [-0.002+0.01, 0, 0.15]))
    # rear line
    p.append(padd(p[3], [0.01, -0.04, -0.15]))
    p.append(padd(p[3], [0.041, 0.02, 0.02]))
    p.append(padd(p[3], [0.025, 0, 0.15]))

    for i, pos in enumerate(p):
        zangle = 0.0
        yangle = 0.0
        if i in [4,5]: # skip these, too much of a pain to remove them
            continue
        if i in [10, 6, 13]:
            yangle = -20.0
            zangle = 3.5
        t.add_joint(pos, yangle=yangle, zangle=zangle)
    door_pts.append(p)

    return door_pts

def door(roof_pts, doorsill_pts, floor_offs):

    t = Group(name="door")
    doorpts = door_pts(t, roof_pts, doorsill_pts, floor_offs)

    # connect inner door pieces
    row = doorpts[0]
    for i in range(5):
        diam = dthick
        if i in [3,4]:
            diam = 0.015
        t.add_pipe(row[i], row[i+1], diam)
    t.add_pipe(row[0], row[2])
    t.add_pipe(row[0], row[3])

    # connect outer door pieces
    row2 = doorpts[1]
    for i in range(3):
        t.add_pipe(row2[i], row2[i+1], dthin)
    # window line
    #t.add_pipe(row2[4], row2[5], dthin)
    #t.add_pipe(row2[5], row2[7], dthin)
    t.add_pipe(row2[10], row2[6], dthin)
    t.add_pipe(row[4], row2[13], dthin)
    t.add_pipe(row[5], row2[6], dthin)
    # bars
    for i in range(3):
        t.add_pipe(row2[8+i], row2[11+i])
    for i in range(2):
        t.add_pipe(row2[8+i+1], row2[11+i], dthin)
        t.add_pipe(row2[8+i], row2[9+i], dthin)
        t.add_pipe(row2[11+i], row2[12+i], dthin)

    t.add_pipe(row2[7], row2[9], dthin)
    t.add_pipe(row2[7], row2[0], dthin)

    # inner to outer, lower section
    t.add_pipe(row[1], row2[11], dthin)
    t.add_pipe(row[1], row2[8], dthin)
    t.add_pipe(row[2], row2[11], dthin)

    # inner to outer, upper section
    t.add_pipe(row2[10], row2[0], dthin)
    t.add_pipe(row2[13], row[3], dthin)
    t.add_pipe(row2[11], row[3], dthin)
    t.add_pipe(row2[8], row[0], dthin)
    return doorpts, t

################
# floor support
################

def create_support_pts(cfloor_pts, fpoints, roof_pts):
    # take each corner, make large runners right the way along
    support = [[], [], []]
    ht = 0.040
    woffs = 0.035
    support[0].append(padd(copy(cfloor_pts[1][0]), [0.015, 0.005, ht]))
    support[0].append(padd(copy(fpoints[1][0]), [0.015, 0.030, ht]))
    support[1].append(padd(copy(cfloor_pts[-2][0]), [-0.025, 0.005, ht]))
    support[1].append(padd(copy(fpoints[2][0]), [-0.025, 0.030, ht]))
    support[0][0][0] = cfloor_pts[1][2][0] - woffs
    support[0][1][0] = cfloor_pts[1][2][0] - woffs
    support[1][0][0] = cfloor_pts[2][2][0] + woffs
    support[1][1][0] = cfloor_pts[2][2][0] + woffs

    # just off of dashboard
    support[2].append(copy(roof_pts[3][0]))
    support[2].append(copy(roof_pts[3][-1]))
    boffs = 0.0235
    support[2][0][1] += boffs
    support[2][1][1] += boffs
    # back of floor
    support[2].append(copy(cfloor_pts[0][0]))
    support[2].append(copy(cfloor_pts[-1][0]))
    # around passenger elbow (nominated "centre")
    support[2].append(copy(cfloor_pts[0][1]))
    support[2].append(copy(cfloor_pts[-1][1]))
    hoffs = 0.30
    boffs = 0.07
    woffs = 0.07
    support[2][4][2] += hoffs
    support[2][5][2] += hoffs
    support[2][4][1] += boffs
    support[2][5][1] += boffs
    support[2][4][0] -= woffs
    support[2][5][0] += woffs
    # front of floor
    support[2].append(copy(cfloor_pts[0][-1]))
    support[2].append(copy(cfloor_pts[-1][-1]))
    boffs = dthick
    support[2][6][1] += boffs
    support[2][7][1] += boffs
    support[2][6][2] += dthick / pow(2.0, 0.5)
    support[2][7][2] += dthick / pow(2.0, 0.5)

    return support

def floor_support(cfloor_pts, fpoints, backbox_pts, roof_pts, floor_offs):

    t = Group(name="floor_support")
    support_pts = create_support_pts(cfloor_pts, fpoints, roof_pts)

    for row in support_pts:
        for pt in row:
            t.add_joint(pt)

    jup4 = [0,0,1.0]
    jdown4 = [0,0,-1.0]
    for row in support_pts[:2]:
        for i in range(len(row)-1):
            t.add_pipe(row[i], row[i+1], dthick, jup, jup)
            t.add_pipe(row[i], row[i+1], dthick)
            t.add_pipe(row[i], row[i+1], dthick, jdown, jdown)

    # along floor side
    t.add_pipe(cfloor_pts[0][0], support_pts[2][6], dthick, jul, jleft0)
    t.add_pipe(cfloor_pts[-1][0], support_pts[2][7], dthick, jur, jright0)
    #support[0].append(padd(cfloor_pts[0][-1], rightup))

    # from door sill up backbox
    m = 1.07
    jfurP = pmul(jfur, m)
    jfulP = pmul(jful, m)
    t.add_pipe(cfloor_pts[0][0], roof_pts[0][-1], dthick, jfurP, jfdr)
    t.add_pipe(cfloor_pts[-1][0], roof_pts[0][0], dthick, jfulP, jfdl)

    # create side-door support
    jfurP = pmul(jfur, m)
    jfulP = pmul(jful, m)
    jfulP[1] *= 1.5
    jfurP[1] *= 1.5
    jfulP[2] /= 1.4
    jfurP[2] /= 1.4
    # centre to dashboard
    t.add_pipe(support_pts[2][0], support_pts[2][4], dthick, jback)
    t.add_pipe(support_pts[2][1], support_pts[2][5], dthick, jback)
    # centre to back floor
    t.add_pipe(support_pts[2][4], support_pts[2][2], dthick, AUTO, jfurP)
    t.add_pipe(support_pts[2][5], support_pts[2][3], dthick, AUTO, jfulP)
    # centre to high roof corner
    jfdP = pmul(jfd, 1.85)
    jfdP[2] /= 2.4
    t.add_pipe(support_pts[2][4], roof_pts[0][-1], dthick, AUTO, jfdP)
    t.add_pipe(support_pts[2][5], roof_pts[0][0], dthick, AUTO, jfdP)
    # centre to front floor
    m = 0.8
    jbuP = pmul(jbu, m)
    #jbuP[1] /= 1.8
    t.add_pipe(support_pts[2][6], support_pts[2][4], dthick, jbuP)
    t.add_pipe(support_pts[2][7], support_pts[2][5], dthick, jbuP)
    # dashboard down to front floor
    t.add_pipe(support_pts[2][6], support_pts[2][0], dthick, AUTO, jdown0)
    t.add_pipe(support_pts[2][7], support_pts[2][1], dthick, AUTO, jdown0)
    # dashboard up to roof corner
    t.add_pipe(roof_pts[2][1], support_pts[2][0], dthick, jdown0, jup0)
    t.add_pipe(roof_pts[2][-2], support_pts[2][1], dthick, jdown0, jup0)

    return support_pts, t

################
# wheels
# https://www.johnbrownwheels.com/product/55-x-13-minilight-4x1143-5513ML4114
# 4x114.3 mm bolt pattern
################

wht = h0 + 0.05
axle_width = 1.4
def wheel_pos(offs):
    return [[axle_width/2.0, front_fwd[2]-0.16, wht],
             [-axle_width/2.0, front_fwd[2]-0.16, wht],
             [axle_width/2.0, offs[1]+back_fwd[2]-0.195, wht],
             [-axle_width/2.0, offs[1]+back_fwd[2]-0.195, wht]
            ]

def wheels(offs):
    wp = wheel_pos(offs)
    t = Group(name="wheels")
    for i, pos in enumerate(wp):
        t.add_wheel(pos, wheel_angle if i in [0,1] else 0.0)
    return t


############
# rear swing arm and frame
############

def rearframe(left=False):
    u = union()
    wid = 0.05
    side = union()
    s = union()
    side.add(translate([-0.03, 0, 0]).add(s))

    # swing arm
    c = cube([wid*2, 0.65, wid*2])
    c = translate([-0.05, 0, -wid]).add(c)
    c = rotate([16.0, 0, 0]).add(c)
    side.add(c)
    # shock
    c = cylinder(r=0.02, h=0.3, center=True, segments=15)
    c = rotate([0, 0, 90]).add(c)
    c = translate([0, 0.0, 0.2]).add(c)
    side.add(c)
    # swingarm mount rod
    c = cylinder(r=0.03, h=0.30, center=True, segments=15)
    c = rotate([0, 90, 0]).add(c)
    c = translate([-0.07, 0.55, 0.270]).add(c)
    side.add(c)
    # swingarm bolt
    c = cylinder(r=0.015, h=0.30, center=True, segments=15)
    c = rotate([0, 90, 0]).add(c)
    c = translate([-0.07, 0.61, 0.175]).add(c)
    side.add(c)
    # shock-holder
    c = cylinder(r=0.04, h=0.07, center=True, segments=15)
    c = rotate([0, 0, 90]).add(c)
    c = translate([0, 0.0, 0.38]).add(c)
    side.add(c)
    # floor piece behind rear passenger
    c = box50100(0.255)
    c = translate([0, 0, -wid]).add(c)
    c = rotate([0.0, 0, 90]).add(c)
    c = translate([0.06, 0.56, 0.06]).add(c)
    side.add(c)
    c = box50100(0.255)
    c = translate([0, 0, -wid]).add(c)
    c = rotate([0.0, 0, 90]).add(c)
    c = translate([0.06, 0.46, 0.06]).add(c)
    side.add(c)

    # long top support for shock-holder
    c = box50100(0.80)
    c = translate([0, 0, -wid]).add(c)
    c = translate([-0.15, -0.19, 0.35]).add(c)
    s.add(c)
    # top joiner down to swingarm pivot
    #c = box50100(0.284)
    #c = translate([0, 0, -wid]).add(c)
    #c = rotate([-15.0, 0, 0]).add(c)
    #c = translate([-0.15, 0.31, 0.36]).add(c)
    #s.add(c)
    # top joiner down to swingarm pivot
    c = box50100(0.24)
    c = translate([0, 0, -wid]).add(c)
    c = rotate([-90.0, 0, 0]).add(c)
    c = translate([-0.15, 0.52, 0.30]).add(c)
    s.add(c)
    # cross-member to centre
    c = box50100(0.375)
    c = translate([0, 0, -wid]).add(c)
    c = rotate([0.0, 0, -90]).add(c)
    c = translate([-0.525, 0.22, 0.35]).add(c)
    s.add(c)
    # side-right to bumper
    c = box50100(0.31)
    c = translate([0, 0, -wid]).add(c)
    c = rotate([90.0, 0, 0]).add(c)
    c = rotate([0.0, 90, 0]).add(c)
    c = translate([-0.15, -0.19, 0.45]).add(c)
    s.add(c)
    # side-right to back box
    c = box50100(0.46)
    c = translate([0, 0, -wid]).add(c)
    c = rotate([90.0, 0, 0]).add(c)
    c = rotate([0.0, 90, 0]).add(c)
    c = translate([-0.28, 0.37, 0.45]).add(c)
    s.add(c)
    # upright to back box
    #c = box50100(0.12)
    #c = translate([0, 0, -wid]).add(c)
    #c = rotate([-90.0, 0, 0]).add(c)
    #c = rotate([0.0, 0, 90]).add(c)
    #c = translate([-0.15, 0.37, 0.49]).add(c)
    #s.add(c)
    # upright for motor
    c = box50100(0.09)
    c = translate([0, 0, -wid]).add(c)
    c = rotate([-90.0, 0, 0]).add(c)
    c = translate([-0.480, 0.17, 0.44]).add(c)
    s.add(c)

    # upright for back support
    c = box50100(0.74)
    c = translate([0, 0, -wid]).add(c)
    c = rotate([0.0, 90, 0]).add(c)
    c = rotate([98.0, 00, 0]).add(c)
    c = translate([0.11, 0.47, 0.09]).add(c)
    side.add(c)

    # long side along door sill
    c = box50100(1.07)
    c = translate([0, 0, -wid]).add(c)
    c = rotate([0.0, 90, 0]).add(c)
    c = translate([0.11, 0.46, 0.11]).add(c)
    side.add(c)
    # short side parallel to long door sill side
    c = box50100(0.24)
    c = translate([0, 0, -wid]).add(c)
    c = rotate([0.0, 90, 0]).add(c)
    c = translate([-0.195, 0.43, 0.06]).add(c)
    side.add(c)

    # foot-well cross-member to centre
    c = box50100(0.615)
    c = translate([0, 0, -wid]).add(c)
    c = rotate([0.0, 90, 90]).add(c)
    c = translate([-0.525, 1.52, 0.11]).add(c)
    s.add(c)

    # long side to front lower wishbone
    c = box50100(1.2)
    c = translate([0, 0, -wid]).add(c)
    c = rotate([0.0, 90, 0]).add(c)
    c = translate([-0.31, 1.52, 0.11]).add(c)
    side.add(c)

    # upright to hold steering
    c = box50100(0.38)
    c = translate([0, 0, -wid]).add(c)
    c = rotate([0.0, 90, 0]).add(c)
    c = rotate([90.0, 00, 0]).add(c)
    c = translate([-0.31, 1.64, 0.11]).add(c)
    side.add(c)

    # upright to front door sill
    c = box50100(0.38)
    c = translate([0, 0, -wid]).add(c)
    c = rotate([0.0, 90, 0]).add(c)
    c = rotate([90.0, 00, 0]).add(c)
    c = translate([0.11, 1.43, 0.11]).add(c)
    side.add(c)

    sep = 0.19 # separation between sides
    if left:
        u.add(translate([sep/2, 0, 0]).add(side))
    else:
        u.add(translate([-sep/2, 0, 0]).add(side))

    wheelhub = cylinder(r=wheel_radius/2, h=0.060, center=True, segments=30)
    # position motor at pivot point
    wheelhub = rotate([0,90,0]).add(wheelhub)
    wheelhub = translate([0, 0.0, 0]).add(wheelhub)
    u.add(wheelhub)
    return u


############
# double wishbone
############

def doublewish(left=False):
    u = union()
    wid = 0.04
    side = union()
    # hub mount
    c = cube([wid, wid*2, 0.20])
    c = translate([-0.12, -0.05, -0.10]).add(c)
    side.add(c)

    springed = 15 # angle of rotation of wishbone

    # lower wishbone
    c = cylinder(r=0.015, h=0.36, center=False, segments=15)
    c = rotate([0, -90, 0]).add(c)
    c = rotate([0, 0, 20]).add(c)
    c = rotate([0, springed, 0]).add(c)
    c = translate([-0.12, -0.05, -0.10]).add(c)
    side.add(c)
    c = cylinder(r=0.015, h=0.36, center=False, segments=15)
    c = rotate([0, -90, 0]).add(c)
    c = rotate([0, 0, -20]).add(c)
    c = rotate([0, springed, 0]).add(c)
    c = translate([-0.12, 0.03, -0.10]).add(c)
    side.add(c)

    # upper wishbone
    c = cylinder(r=0.015, h=0.32, center=False, segments=15)
    c = rotate([0, -90, 0]).add(c)
    c = rotate([0, 0, 15]).add(c)
    c = rotate([0, springed, 0]).add(c)
    c = translate([-0.12, -0.05, 0.10]).add(c)
    side.add(c)
    c = cylinder(r=0.015, h=0.32, center=False, segments=15)
    c = rotate([0, -90, 0]).add(c)
    c = rotate([0, 0, -15]).add(c)
    c = rotate([0, springed, 0]).add(c)
    c = translate([-0.12, 0.03, 0.10]).add(c)
    side.add(c)

    u.add(translate([0, 0, 0]).add(side))

    wheelhub = cylinder(r=wheel_radius/2, h=0.020, center=True, segments=30)
    # position motor at pivot point
    wheelhub = rotate([0,90,0]).add(wheelhub)
    wheelhub = translate([0, 0.0, 0]).add(wheelhub)
    u.add(wheelhub)
    return u


############
# rear_bumper
############

def rbumper_pts(t, offs, bump_rear_ht):
    # base of bonnet
    base = []

    wpout = wpoints[-1]-0.05 # bring sides of back bumper in a bit

    # top line
    wp1 = wpoints[-2]-0.16
    wp0 = wpoints[1]-0.04
    wdiff = wp1 - wp0
    p = []
    for j, wd in enumerate([-wp1, -wp0, wp0, wp1]):
        pos = [wd+offs[0], back_fwd[-1]+offs[1], h2-0.04+offs[2]]
        p.append(pos)
        t.add_joint(pos)
    base.append(p)

    # middle rows
    p1 = []
    wp2 = wpoints[-1] - wdiff
    for i, d in enumerate([back_fwd[-1]-0.03, back_fwd[-2]]):
        p = []
        for j, wd in enumerate([-wpout, -wp2, wp2, wpout]):
            height = h1-0.04
            if i == 0:
                height += 0.04 # raise spoiler line slightly
                wd -= math.copysign(0.03, wd) # bring rear bumper in
            # bring inner part in too, to make room for metal frame
            if i == 1 and j in [1,2]:
                wd -= math.copysign(0.04, wd)
            pos = [wd+offs[0], d+offs[1], height+offs[2]]
            p.append(pos)
            t.add_joint(pos)
        p1.append(p)
    base.append(p1)

    # bottom row, raised up above wheel centres (clearance for swingarm)
    p = []
    wp2 -= 0.05
    d = back_fwd[-1]-0.1
    for j, wd in enumerate([-wpoints[-2], -wp2, wp2, wpoints[-2]]):
        pos = [wd+offs[0], d+offs[1], bump_rear_ht+offs[2]+0.15]
        p.append(pos)
        t.add_joint(pos)
    base.append(p)

    return base

def rear_bumper(offs, bump_rear_ht):
    t = Group(name="rear_bumper")
    base = rbumper_pts(t, offs, bump_rear_ht)

    def connect_straight(p, thickness=dthin):
        t.add_pipe(p[0], p[1], thickness)
        t.add_pipe(p[3], p[2], thickness)
    def connect_rows(p, p1, thickness=dthin):
        for j in range(len(p)):
            offs1, offs2 = AUTO, AUTO
            if p == base[0] and p1 == base[1][1]:
                offs1 = [jlfrot4, jfd, jfd, jrfrot4][j]
            t.add_pipe(p[j], p1[j], thickness, offs1, offs2)
        connect_diags(p, p1)
    def connect_diags(p, p1):
        loffs1, loffs2 = AUTO, AUTO
        roffs1, roffs2 = AUTO, AUTO
        if p == base[2] and p1 == base[1][1]:
            loffs1 = jrf
            roffs1 = jlf
            loffs2 = jlbdrotz2
            roffs2 = jrbdrotz2
        t.add_pipe(p[0], p1[1], dthin, loffs1, loffs2)
        t.add_pipe(p[3], p1[2], dthin, roffs1, roffs2)
    # connect row-to-row
    connect_rows(base[0], base[1][0], dthick)
    connect_rows(base[2], base[1][0], dthick)
    connect_rows(base[2], base[1][1], dthick)
    connect_rows(base[0], base[1][1], dthick)
    connect_rows(base[1][0], base[1][1], dthick)
    connect_straight(base[0], dthick)
    connect_straight(base[2], dthick)
    connect_straight(base[1][0], dthick)
    connect_straight(base[1][1])
    # lastly cross-bumper, top line
    t.add_pipe(base[0][1], base[0][2], dthick)

    return base, t

#################
# windshield
#################

def windshield(roofpts, airbox_pts, thick):

    rbase, rmid, rscreen, doorpts = roofpts
    aback, atop = airbox_pts

    p0 = rscreen
    p1 = []
    p2 = []
    p3 = [doorpts[1], atop[0][-1], atop[1][-1], doorpts[2]]
    for i in range(len(rscreen)):
        p = pavg(p0[i], p3[i], 0.8)
        if i in [1, 2]:
            p[2] += 0.06
        p1.append(p)
        p = pavg(p0[i], p3[i], 0.2)
        if i in [1, 2]:
            p[2] += 0.06
        p2.append(p)
    plist = [p0, p1, p2, p3]

    if spline_me: plist = spline_double_divide(plist, step_v=0.18, step_h=.18)

    po = thick
    return SurfaceFromCurves(plist).create_surface(po, (1.0, 1.0, 1.0))

#################
# roof panel
#################

def roof_panel(roofpts, bbox, rearpts, bumper_pts, thick):

    rbase, rmid, rscreen, doorpts = roofpts
    [roof, boxbase, midbox, midfront] = bbox

    # front needs "construction" - must have same number of points as rest
    rfront = []
    for i in range(len(rscreen)):
        rfront.append(rscreen[3-i])
        p = copy(rscreen[3-i])
        p[0] -= math.copysign(0.001, p[0]) # adjust point in slightly
        rfront.append(p)

    # minor weirdness, here, constructing the panel due to staggered triangles
    p0 = copy(rearpts[2])
    p0[0] -= 0.001
    p1 = copy(rearpts[-1])
    p1[0] += 0.001
    rp0 = copy(rearpts[3])
    rp0[0] -= 0.001
    rp1 = copy(rearpts[1])
    rp1[0] += 0.001
    rearpanel = rearpts[:2] + [rp1, p0, rp0] + \
                 rearpts[3:] + [p1]
    rearpanel.reverse()

    print "pts", len(rearpanel), len(roof), rearpanel
    plist = [rfront, rmid, rbase, roof, rearpanel]

    po = thick
    return SurfaceFromCurves(plist).create_surface(po, (1.0, 1.0, 1.0))

#################
# cockpit floor
#################

def cockpit_floor_panel(cfloor_pts, fpoints, bbox, thick):

    [roof, boxbase, midbox, midfront] = bbox
    fp = []
    for i in range(len(fpoints)):
        fp.append(fpoints[i][0])
    plist = [boxbase] + map(list, zip(*cfloor_pts)) + [fp]

    po = thick
    return SurfaceFromCurves(plist).create_surface(po, (1.0, 1.0, 1.0))

#################
# bonnet
#################

def bonnet_panel(bpoints, fpoints, thick):

    bbase, bscoop = bpoints

    plist = []
    for i in range(3):
        p = bbase[i]
        plist.append([bbase[0][i], bscoop[0][i], bscoop[1][i], bbase[-1][i]])

    if False:
        # extra work to create front merged panel line
        p0 = copy(bbase[0][2])
        p0[1] -= 0.001
        p1 = copy(bbase[-1][2])
        p1[1] -= 0.001
        for i in range(len(bbase)):
            print len(bbase[i])
        plist.append([p0, bbase[1][3], bbase[2][3], p1])

        # front rear line
        fline = []
        for i in range(len(fpoints)):
            fline.append(fpoints[i][-1])
        plist.append(fline)

    po = thick
    return SurfaceFromCurves(plist).create_surface(po, (1.0, 1.0, 1.0))

###############
# car
###############

# helps implicitcad (no support for color)
def _color(c):
    return union()

def create_car(t, _scl, _mirror, colour):

    # bonnet
    c = color([0.0, 0.6, 0.2])
    b = Bonnet()
    b.visible = False
    bpoints, b = b.make()
    bbase, bscoop = bpoints
    t.add(b, c)

    # front section
    #c = color([0.9, 0.3, 0.2, 1.0])
    c = color([1.0, 0.65, 0.0, 1.0])
    f = Front()
    f.visible = False
    fpoints, f = f.make(bpoints)
    t.add(f, c)

    # cockpit floor
    c = color([0.8, 0.9, 0.2, 1.0])
    floor_offs = [0, front_fwd[-1], 0]
    f = Floor()
    f.visible = False
    cfloor_pts, _ = f.make(fpoints, floor_offs)
    t.add(f, c)

    # central airbox
    c = color([0.2, 0.3, 0.9, 1.0])
    f = Airbox()
    #f.visible = False
    airbox_pts, f = f.make(bpoints, cfloor_pts, fpoints, floor_offs)
    t.add(f, c)

    # roof
    c = color([0.7, 0.3, 0.9, 0.3])
    cfloor_offs = [0, floor_offs[1] + middle_fwd[-1], 0]
    f = Roof()
    f.visible = False
    roof_pts, _ = f.make(airbox_pts, bscoop, cfloor_offs)
    t.add(f, c)

    # dashboard and door-holder
    c = color([1.0, 0.75, 0.79, 1.0])
    f = dashboard_door(airbox_pts, bpoints, roof_pts,
                                 cfloor_pts, fpoints, floor_offs)
    #f.visible = False
    t.add(f, c)

    # back box behind door
    c = color([0.7, 0.5, 0.1, 1.0])
    back_offs = [0, cfloor_offs[1] + back_fwd[-1], 0]
    f = BackBox()
    #f.visible = False
    bump_rear_ht = h0#+0.20
    bpts, _ = f.make(roof_pts, airbox_pts, cfloor_pts, back_offs, bump_rear_ht)
    t.add(f, c)

    # wheels
    c = color([0.7, 0.0, 0.1, 0.6])
    f = wheels(cfloor_offs)
    f.visible = False
    t.add(f, c)

    # rear bumper
    c = color([0.9, 0.3, 0.2, 1.0])
    rbump, f = rear_bumper(cfloor_offs, bump_rear_ht)
    f.visible = False
    t.add(f, c)

    # rear section
    c = color([0.1, 0.5, 0.0, 1.0])
    f = RearSection()
    f.visible = False
    rear, _ = f.make(rbump, bpts, cfloor_offs)
    t.add(f, c)

    # door sill
    c = color([0.2, 0.6, 0.0, 1.0])
    pdoorsill, f = door_sill(airbox_pts, bpoints, roof_pts,
                                 cfloor_pts, fpoints, bpts, floor_offs)
    f.visible = False
    t.add(f, c)

    # floor support
    c = color([1.0, 0.2, 0.4, 0.1])
    f_support_pts, f = floor_support(cfloor_pts, fpoints,
                                     bpts, roof_pts, floor_offs)
    #f.visible = False
    t.add(f, c)

    # door angle
    door_angle = 00

    # door
    c = color([0.9, 0.2, 0.3, 1.0])
    pdoors, f = door(roof_pts, pdoorsill, floor_offs)
    woff = wpoints[-1]-0.05
    doff = floor_offs[1]+0.3
    # tiny positioning-box
    c.add(translate([woff,doff,0]).add(cube([0.04, 0.04, 0.04], center=True)))
    woff = wpoints[-1]+0.025
    doff = floor_offs[1]+0.225
    # tiny positioning-box
    c.add(translate([woff,doff,0]).add(cube([0.04, 0.04, 0.04], center=True)))
    transops = [translate([-woff,-doff,0]),
                rotate([0,0,-door_angle]),
                translate([woff,doff,0])]
    t.add(f, c, transops)
    f.visible = False

    return t

    # metre rule
    c = color([0.4, 0.7, 0.8, 1.0])
    t.extras.add(c)
    if True:
        p = cylinder(r=0.01, h=1.0, segments=20)
        c.add(translate([0, 2.2, 0]).add(p))
        c.add(translate([0, 0.0, 0]).add(p))
        c.add(translate([0, 3.4, 0]).add(p))

        p = cylinder(r=0.01, h=1.0, center=True, segments=20)
        p = rotate([90,0,0]).add(p)
        c.add(translate([0.9, 1.8, 0.5]).add(p))

    # rear frame measure sticks
    if True:
        # top view
        p = cylinder(r=0.004, h=0.87, center=False, segments=20)
        p = rotate([90,0,0]).add(p)
        c.add(translate([0.53, 3.145, 0.35]).add(p))
        p = cylinder(r=0.004, h=0.525, center=False, segments=20)
        p = rotate([90,0,0]).add(p)
        c.add(translate([0.40, 3.145, 0.52]).add(p))
        p = cylinder(r=0.004, h=0.61, center=False, segments=20)
        p = rotate([90,0,0]).add(p)
        p = rotate([0,0,-90]).add(p)
        c.add(translate([0.305, 2.39, 0.05]).add(p))
        p = cylinder(r=0.004, h=1.29, center=False, segments=20)
        p = rotate([90,0,0]).add(p)
        p = rotate([0,0,-90]).add(p)
        c.add(translate([0.645, 2.26, 0.12]).add(p))
        # width check (1.4 metre)
        p = cylinder(r=0.004, h=1.4, center=True, segments=20)
        p = rotate([90,0,0]).add(p)
        p = rotate([0,0,-90]).add(p)
        c.add(translate([0.0, 2.82, 0.0]).add(p))

        # side view
        p = cylinder(r=0.004, h=0.14, center=False, segments=20)
        c.add(translate([0.41, 3.16, 0.36]).add(p))
        p = cylinder(r=0.004, h=0.44, center=False, segments=20)
        c.add(translate([0.41, 2.5, 0.06]).add(p))

    # get four wheel positions
    wp = wheel_pos(cfloor_offs)

    if True:
        # add swingarm
        a = rearframe()
        a = rotate([0,0,180]).add(a)
        a = translate(wp[3]).add(a)
        a = translate([0.05,0,0]).add(a)
        t.extras.add(a)
        a = rearframe()
        a = rotate([0,0,180]).add(a)
        a = mirror([-1,0,0]).add(a)
        a = translate(wp[2]).add(a)
        a = translate([-0.05,0,0]).add(a)
        t.extras.add(a)

    # front frame measure sticks
    if True:
        # width check (1.4 metre)
        p = cylinder(r=0.004, h=1.4, center=True, segments=20)
        p = rotate([90,0,0]).add(p)
        p = rotate([0,0,-90]).add(p)
        c.add(translate([0.0, 0.35, 0.0]).add(p))

    if True:
        # add right doublewishbone
        a = doublewish()
        a = rotate([0,0,180]).add(a)
        a = translate(wp[1]).add(a)
        a = translate([0.03,0,0]).add(a)
        t.extras.add(a)

    if True:
        # generator
        c = color([0.8, 1.0, 0.8, 0.6])
        woff = 0.0
        doff = 0.39
        c.add(translate([woff,doff,0.3]).add(cube([0.38, 0.45, 0.58], center=True)))
        t.extras.add(c)
        #f.visible = False

        # battery
        c = color([0.8, 0.8, 1.0, 0.6])
        woff = 0.0
        doff = 0.85 # 2.30
        hoff = 0.42 # 0.14
        c.add(translate([woff,doff,hoff]).add(cube([0.500, 0.400, 0.200], center=True)))
        t.extras.add(c)
        #f.visible = False

    return
    if False:
        # windshield
        w = windshield(roof_pts, airbox_pts, thick=0.01)
        c = color([0.4, 0.7, 0.8, 0.1])
        c.add(w)
        t.extras.add(c)

    # cockpit floor
    w = cockpit_floor_panel(cfloor_pts, fpoints, bpts, thick=0.01)
    c = color([0.4, 0.7, 0.3, 0.2])
    c.add(w)
    t.extras.add(c)

    if False:
        # roof_panel
        w = roof_panel(roof_pts, bpts, rear, rbump, thick=0.01)
        c = color([0.9, 0.55, 0.1, 0.4])
        c.add(w)
        t.extras.add(c)

    # bonnet
    w = bonnet_panel(bpoints, fpoints, thick=0.01)
    c = color([0.9, 0.55, 0.1, 0.4])
    c.add(w)
    t.extras.add(c)


if __name__ == '__main__':

    scl = 1.0
    t = Model()
    c = [0.6, 0.5, 0.8, 1.0]
    c = True
    create_car(t, scl, 1.0, c)

    m = t.get_model()
    print m

    scad_render_to_file(m, "car_prog2.scad")
    f = open("car_model.scad", "w")
    f.write("""include<car_prog2.scad>;
//%import_stl(filename = "car_model.stl");
    """)
    f.close()

    print "num joints:", len(t.joints)
    print "num pipes:", len(t.pipes)
    print "pipe len (m):", t.plen
